"""
win_pyxs.gplpv contains an implementation of a pyxs.connection.PacketConnection
which sends the packets over the PCI device exposed by the GPLPV drivers for
Windows.
"""

from __future__ import print_function

__all__ = ['XenBusConnectionGPLPV']

import ctypes
from ctypes.wintypes import HANDLE
from ctypes.wintypes import BOOL
from ctypes.wintypes import HWND
from ctypes.wintypes import DWORD
from ctypes.wintypes import WORD
from ctypes.wintypes import LONG
from ctypes.wintypes import ULONG
from ctypes.wintypes import LPCSTR
from ctypes.wintypes import HKEY
from ctypes.wintypes import BYTE
import logging
import socket

import backports.socketpair
import six
from win32file import CreateFile, CloseHandle, ReadFile, WriteFile
from win32file import (
    FILE_GENERIC_READ, FILE_GENERIC_WRITE, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL
)

import pyxs.connection
from pyxs._internal import NUL

from .exceptions import GPLPVDeviceOpenError, GPLPVDriverError

_winDevicePath = None


class XenBusConnectionGPLPV(pyxs.connection.PacketConnection):
    """
    A pyxs.PacketConnection which communicates with xenstore over the PCI
    device exposed by the GPLPV drivers on Windows. The interface of this
    driver is very similar to the ones on Linux (direct reads/writes to a
    file-like object) so we reuse most of the PacketConnection class and leave
    the implementation detail to the XenBusTransportGPLPV.
    """

    def create_transport(self):
        """
        Initialises a new instance of XenBusTransportGPLPV to communicate with
        xenstore using the GPLPV drivers on Windows.
        """
        return XenBusTransportGPLPV()

    def recv(self):
        """
        This version of recv simply wraps the version from the superclass to
        ensure that a second recv on the XenBusTransportGPLPV is performed
        even for packets where no data actually needed to be read from
        xenstore. Because the read is of 0 length no actual access to the
        device will be performed but the sends and recvs will tally up
        correctly which is important to ensure we do not cause a deadlock when
        reading the file.
        """
        packet = super(XenBusConnectionGPLPV, self).recv()
        if not packet.payload:
            self.transport.recv(0)
        return packet


class XenBusTransportGPLPV(object):
    """
    A transport for pyxs which communicates with xenstore using the PCI device
    exposed by the GPLPV drivers for Windows. The PCI device is a file-like
    object which can only be written to/read from using Windows APIs. Despite
    the alternate APIs the goal is to keep things as similar as possible to the
    equivalent transports in pyxs. The major limitation preventing this is that
    it is not possible to use select on the file-like objects used here (as
    select on Windows can only be used on sockets) so a less-optimal way
    involving a socketpair has been used instead.
    """

    def __init__(self):
        global _winDevicePath

        self._logger = logging.getLogger(
            __name__ + '.' + self.__class__.__name__
        )

        self.fd = None
        self.notify = False

        # A socket pair which can be used to mimic the default pyxs behaviour
        # of returning a fileno which can be slected on to check when data is
        # available
        self.r_terminator, self.w_terminator = socket.socketpair()

        # Once the windows device path is learned once reuse it otherwise
        # ctypes.POINTER() for the same structure leaks memory. Although
        # this can be reclaimed with ctypes._reset_cache() this is poking
        # at the internals of ctypes which doesn't seem to be a good idea.
        if _winDevicePath:
            self.path = _winDevicePath
            self._open_device()
            return

        # Determine self.path using some magic Windows code which is derived
        # from:
        #   http://pydoc.net/Python/pyserial/2.6/serial.tools.list_ports_windows/.
        # The equivalent C from The GPLPV driver source can be found in
        # get_xen_interface_path() of shutdownmon:
        #   http://xenbits.xensource.com/ext/win-pvdrivers/file/896402519f15/shutdownmon/shutdownmon.c

        DIGCF_PRESENT = 2
        DIGCF_DEVICEINTERFACE = 16
        NULL = None
        ERROR_SUCCESS = 0
        ERROR_INSUFFICIENT_BUFFER = 122
        ERROR_NO_MORE_ITEMS = 259

        HDEVINFO = ctypes.c_void_p
        PCTSTR = ctypes.c_char_p
        CHAR = ctypes.c_char
        PDWORD = ctypes.POINTER(DWORD)
        LPDWORD = ctypes.POINTER(DWORD)
        PULONG = ctypes.POINTER(ULONG)

        # Return code checkers
        def ValidHandle(value, func, arguments):
            if value == 0:
                raise GPLPVDriverError(str(ctypes.WinError()))
            return value

        # Some structures used by the Windows API
        class GUID(ctypes.Structure):
            _fields_ = [
                ('Data1', DWORD),
                ('Data2', WORD),
                ('Data3', WORD),
                ('Data4', BYTE * 8),
            ]

            def __str__(self):
                return "{%08x-%04x-%04x-%s-%s}" % (
                    self.Data1,
                    self.Data2,
                    self.Data3,
                    ''.join(["%02x" % d for d in self.Data4[:2]]),
                    ''.join(["%02x" % d for d in self.Data4[2:]]),
                )

        PGUID = ctypes.POINTER(GUID)

        class SP_DEVINFO_DATA(ctypes.Structure):
            _fields_ = [
                ('cbSize', DWORD),
                ('ClassGuid', GUID),
                ('DevInst', DWORD),
                ('Reserved', PULONG),
            ]

            def __str__(self):
                return "ClassGuid:%s DevInst:%s" % (
                    self.ClassGuid, self.DevInst
                )

        PSP_DEVINFO_DATA = ctypes.POINTER(SP_DEVINFO_DATA)

        class SP_DEVICE_INTERFACE_DATA(ctypes.Structure):
            _fields_ = [
                ('cbSize', DWORD),
                ('InterfaceClassGuid', GUID),
                ('Flags', DWORD),
                ('Reserved', PULONG),
            ]

            def __str__(self):
                return "InterfaceClassGuid:%s Flags:%s" % (
                    self.InterfaceClassGuid, self.Flags
                )

        PSP_DEVICE_INTERFACE_DATA = ctypes.POINTER(SP_DEVICE_INTERFACE_DATA)
        PSP_DEVICE_INTERFACE_DETAIL_DATA = ctypes.c_void_p

        # Import the Windows APIs
        setupapi = ctypes.windll.LoadLibrary("setupapi")

        SetupDiGetClassDevs = setupapi.SetupDiGetClassDevsA
        SetupDiGetClassDevs.argtypes = [PGUID, PCTSTR, HWND, DWORD]
        SetupDiGetClassDevs.restype = HDEVINFO
        SetupDiGetClassDevs.errcheck = ValidHandle

        SetupDiEnumDeviceInterfaces = setupapi.SetupDiEnumDeviceInterfaces
        SetupDiEnumDeviceInterfaces.argtypes = [
            HDEVINFO, PSP_DEVINFO_DATA, PGUID, DWORD, PSP_DEVICE_INTERFACE_DATA
        ]
        SetupDiEnumDeviceInterfaces.restype = BOOL

        SetupDiGetDeviceInterfaceDetail = \
            setupapi.SetupDiGetDeviceInterfaceDetailA
        SetupDiGetDeviceInterfaceDetail.argtypes = [
            HDEVINFO, PSP_DEVICE_INTERFACE_DATA,
            PSP_DEVICE_INTERFACE_DETAIL_DATA, DWORD, PDWORD, PSP_DEVINFO_DATA
        ]
        SetupDiGetDeviceInterfaceDetail.restype = BOOL

        SetupDiDestroyDeviceInfoList = setupapi.SetupDiDestroyDeviceInfoList
        SetupDiDestroyDeviceInfoList.argtypes = [HDEVINFO]
        SetupDiDestroyDeviceInfoList.restype = BOOL

        # Do stuff
        GUID_XENBUS_IFACE = GUID(
            0x14ce175a, 0x3ee2, 0x4fae,
            (BYTE * 8)(0x92, 0x52, 0x0, 0xdb, 0xd8, 0x4f, 0x1, 0x8e)
        )

        handle = SetupDiGetClassDevs(
            ctypes.byref(GUID_XENBUS_IFACE), NULL, NULL,
            DIGCF_PRESENT | DIGCF_DEVICEINTERFACE
        )

        sdid = SP_DEVICE_INTERFACE_DATA()
        sdid.cbSize = ctypes.sizeof(sdid)
        if not SetupDiEnumDeviceInterfaces(
            handle, NULL, ctypes.byref(GUID_XENBUS_IFACE), 0,
            ctypes.byref(sdid)
        ):
            if ctypes.GetLastError() != ERROR_NO_MORE_ITEMS:
                raise GPLPVDriverError(str(ctypes.WinError()))

        buf_len = DWORD()
        if not SetupDiGetDeviceInterfaceDetail(
            handle, ctypes.byref(sdid), NULL, 0, ctypes.byref(buf_len), NULL
        ):
            if ctypes.GetLastError() != ERROR_INSUFFICIENT_BUFFER:
                raise GPLPVDriverError(str(ctypes.WinError()))

        # We didn't know how big to make the structure until buf_len is
        # assigned...
        class SP_DEVICE_INTERFACE_DETAIL_DATA_A(ctypes.Structure):
            _fields_ = [
                ('cbSize', DWORD),
                ('DevicePath', CHAR * (buf_len.value - ctypes.sizeof(DWORD))),
            ]

            def __str__(self):
                return "DevicePath:%s" % (self.DevicePath, )

        sdidd = SP_DEVICE_INTERFACE_DETAIL_DATA_A()
        sdidd.cbSize = ctypes.sizeof(
            ctypes.POINTER(SP_DEVICE_INTERFACE_DETAIL_DATA_A)
        )
        if not SetupDiGetDeviceInterfaceDetail(
            handle, ctypes.byref(sdid), ctypes.byref(sdidd), buf_len, NULL,
            NULL
        ):
            raise GPLPVDriverError(str(ctypes.WinError()))

        self.path = "" + sdidd.DevicePath

        SetupDiDestroyDeviceInfoList(handle)

        _winDevicePath = self.path

        self._open_device()

    def _open_device(self):
        try:
            self.fd = CreateFile(
                self.path, FILE_GENERIC_READ | FILE_GENERIC_WRITE, 0, None,
                OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, None
            )
        except Exception as exc:
            self._logger.exception('Exception opening GPLPV device:')
            six.raise_from(
                GPLPVDeviceOpenError(
                    "Error while opening {0!r}".format(self.path)
                ), exc
            )

    def fileno(self):
        return self.r_terminator.fileno()

    def close(self, silent=True):
        CloseHandle(self.fd)
        self.fd = None

        self.r_terminator.shutdown(socket.SHUT_RDWR)

        self.r_terminator.close()
        self.w_terminator.close()

    def recv(self, size):
        self._logger.debug('recv: %d', size)

        chunks = []
        while size:
            (err, read) = ReadFile(self.fd, size, None)
            if err:
                raise OSError(err)

            chunks.append(read)
            size -= len(read)

        received = 0
        while received < 1:
            data = self.r_terminator.recv(1)
            received += len(data)
        self._logger.debug('recv: read 1 byte from socket, returning data')

        return b"".join(chunks)

    def send(self, data):
        self._logger.debug('send: %d', len(data))

        size = len(data)
        while size:
            err, lwrite = WriteFile(self.fd, data[-size:], None)
            if err:
                raise OSError(err)

            size -= lwrite

        if self.notify:
            self._logger.debug('send: notifying router')
            self.w_terminator.sendall(NUL + NUL)

        self.notify = not self.notify



"""
A sub-module to hold exceptions for win_pyxs. These exceptions all inherit from
pyxs.PyXSError so that they will be caught along with normal exceptions from
pyxs.
"""

__all__ = [
    'WinPyXSError', 'UnknownSessionError', 'GPLPVDeviceOpenError',
    'GPLPVDriverError',
]

from pyxs import PyXSError


class WinPyXSError(PyXSError):
    """
    Base class for all exceptions raised by this module. Derived from
    pyxs.PyXSError so that it is caught in the same way as the exceptions for
    the non-Windows classes.
    """


class UnknownSessionError(WinPyXSError):
    """
    Exception raised when the xenstore session cannot be found. This can happen
    if something removes the session (via the WMI interface) while your program
    is running.
    """


class GPLPVDeviceOpenError(WinPyXSError):
    """
    Exception raised by the GPLPV connection when opening the device fails.
    """


class GPLPVDriverError(WinPyXSError):
    """
    Exception raised by the GPLPV connection when it fails to learn the device
    path used by the driver.
    """



"""
Performs a simple access to xenstore and prints some details about the
current VM.
"""

from __future__ import print_function

import logging
from pprint import pprint

import six

import pyxs

from win_pyxs import XenBusConnectionWinPV, XenBusConnectionGPLPV
from win_pyxs.exceptions import GPLPVDeviceOpenError, GPLPVDriverError


def _basic_logger_init(logger, verbose=False):
    """
    Setup a logger to output to the stderr stream.

    :param logger: The logger to setup.
    :param verbose: If True the logger & formatter will be set to DEBUG level.
    """
    handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(name)s: %(message)s')
    handler.setFormatter(formatter)

    log_level = logging.INFO
    if verbose:
        log_level = logging.DEBUG

    logger.setLevel(log_level)
    handler.setLevel(log_level)

    logger.addHandler(handler)


def _main():
    logger = logging.getLogger('win_pyxs')
    _basic_logger_init(logger, verbose=True)

    try:
        con = XenBusConnectionGPLPV()
        logger.info('Using XenBusConnectionGPLPV')
    except (GPLPVDeviceOpenError, GPLPVDriverError) as gplpv_exc:
        try:
            con = XenBusConnectionWinPV()
            logger.info('Using XenBusConnectionWinPV')
        except Exception as winpv_exc:
            six.raise_from(winpv_exc, gplpv_exc)

    router = pyxs.Router(con)
    with pyxs.Client(router=router) as client:
        my_uuid = client.read("vm")
        print('My UUID: ', my_uuid)
        my_domid = client.read("domid")
        print('My DomID:', my_domid)
        my_mac = client.read("device/vif/0/mac")
        print('My MAC:  ', my_mac)
        caption, first = 'Drivers: ', True
        for driver in client.list("drivers"):
            if first:
                first = False
            print(caption, client.read(driver))
            caption = '         '
        if first:
            print(caption, 'GPLPV (None in xenstore)')
        print('My Home:')
        pprint(client.list("/local/domain/{}".format(my_domid)))


if __name__ == "__main__":
    _main()



"""
win_pyxs.winpv contains an implementation of a pyxs.connection.PacketConnection
which sends the packets over the WMI interface exposed by the WinPV drivers.
Not all xenstore functions are exposed using this WMI interface so this module
is only able to implement those which are available.
"""

from __future__ import print_function

__all__ = ['XenBusConnectionWinPV']

import socket
import sys
from time import sleep

try:
    from Queue import Queue
except ImportError:
    from queue import Queue

import backports.socketpair
import pythoncom
import six
import wmi

import pyxs
import pyxs.connection
from pyxs._internal import Op, Packet, NUL

from .exceptions import UnknownSessionError

sys.coinit_flags = 0

WMI_CONNECT_RETRY_DELAY = 2
WMI_QUERY_RETRY_DELAY = 0.5


class XenBusConnectionWinPV(pyxs.connection.PacketConnection):
    """
    An implementation of a pyxs connection which uses the WMI interface
    provided by the WinPV drivers to communicate with the xenstore.
    """

    def __init__(self, xs_session_name="PyxsSession"):
        super(XenBusConnectionWinPV, self).__init__()

        self.session = None
        self.session_id = None
        self.session_name = xs_session_name

        self.response_packets = Queue()

        # A socket pair which can be used to mimic the default pyxs behaviour
        # of returning a fileno which can be slected on to check when data is
        # available
        self.r_terminator, self.w_terminator = socket.socketpair()

    def _get_xenstore_session(self, wmi_connect_retry=20):
        # Create a WMI Session
        try:
            wmi_session = wmi.WMI(moniker="//./root/wmi", find_classes=False)
            xenstore_base = wmi_session.XenProjectXenStoreBase()[0]
        except Exception as exc:  # WMI can raise all sorts of exceptions
            if wmi_connect_retry > 0:
                sleep(WMI_CONNECT_RETRY_DELAY)
                return self._get_xenstore_session(
                    wmi_connect_retry=(wmi_connect_retry - 1)
                )

            six.raise_from(
                pyxs.PyXSError("Initialising WMI connection failed"), exc
            )

        if self.session_id is None:
            self.session_id = xenstore_base.AddSession(Id=self.session_name)[0]

        wmi_query = (
            "select * from XenProjectXenStoreSession where SessionId"
            " = {id}"
        ).format(id=self.session_id)

        try:
            sessions = wmi_session.query(wmi_query)
        except Exception:
            sleep(WMI_QUERY_RETRY_DELAY)
            try:
                sessions = wmi_session.query(wmi_query)
            except Exception as exc:
                six.raise_from(
                    pyxs.PyXSError("Unable to query for WMI session"), exc
                )

        try:
            return sessions[0]
        except IndexError:
            raise UnknownSessionError(
                "No session with SessionId={}".format(self.session_id)
            )

    def __copy__(self):
        return self.__class__()

    @property
    def is_connected(self):
        """
        Return whether this connection is currently active & connected to
        xenstore via the XenProjectXenStoreSession.
        """
        return self.session is not None

    def fileno(self):
        """
        Return the fileno from the reader half of the socket pair. This way
        when selecting on this object it will return when data becomes
        available.
        """
        return self.r_terminator.fileno()

    def connect(self, wmi_connect_retry=20):
        """
        Connect the WMI session ready for commands to be sent using this
        connection. Because there can be connection issues here this method
        will retry the connection by default. This can be disabled by passing
        wmi_connect_retry=0 and the number of retries is configurable through
        this parameter.
        """
        if self.is_connected:
            return

        if not self.response_packets:
            self.response_packets = Queue()

        if not self.r_terminator or not self.w_terminator:
            self.r_terminator, self.w_terminator = socket.socketpair()

        self.session = self._get_xenstore_session(
            wmi_connect_retry=wmi_connect_retry
        )

    def send(self, packet):
        """
        Emulates sending a packet to xenstore by calling the equivalent WMI
        method on the XenProjectXenStoreSession WMI object. Only a few
        operations (READ, WRITE, RM, DIRECTORY) because that is all that is
        available using the WMI interface. Because the result of the WMI call
        is the equivalent of the packet received from xenstore in the Linux
        device/socket code this method stores it in a FIFO queue for later
        to be returned by the recv() method.
        """
        try:
            if not self.session:
                self.connect()
        except wmi.x_wmi as exc:
            six.raise_from(pyxs.PyXSError, exc)

        if packet.op == Op.READ:
            try:
                result = self.session.GetValue(packet.payload)[0]
            except wmi.x_wmi as exc:
                six.raise_from(
                    pyxs.PyXSError("session.GetValue call failed"), exc
                )
        elif packet.op == Op.WRITE:
            payload = packet.payload.split('\x00', 1)

            try:
                self.session.SetValue(payload[0], payload[1])
            except wmi.x_wmi as exc:
                six.raise_from(
                    pyxs.PyXSError("session.SetValue call failed"), exc
                )

            result = "OK"
        elif packet.op == Op.RM:
            try:
                self.session.RemoveValue(packet.payload)[0]
            except wmi.x_wmi as exc:
                six.raise_from(
                    pyxs.PyXSError("session.RemoveValue call failed"), exc
                )

            result = "OK"
        elif packet.op == Op.DIRECTORY:
            try:
                result = self.session.GetChildren(packet.payload)[0].childNodes
            except wmi.x_wmi as exc:
                six.raise_from(
                    pyxs.PyXSError("session.GetChildren call failed"), exc
                )

            result = "\x00".join(result)
        else:
            raise Exception(
                "Unsupported XenStore Action ({x})".format(x=packet.op)
            )

        self.response_packets.put(
            Packet(packet.op, result, packet.rq_id, packet.tx_id)
        )

        # Notify that data is available
        self.w_terminator.sendall(NUL)

    def recv(self):
        """
        Receive a packet from xenstore. This method does very little because
        the send method is forced to receive the response (it is the result of
        the WMI call) so it is already written to a queue for this method to
        read.
        """
        self.r_terminator.recv(1)
        return self.response_packets.get(False)

    def close(self, silent=True):
        """
        Close the sockets used to notify pyxs when data is ready & cleanup the
        WMI session used to query xenstore.
        """
        pythoncom.CoInitialize()
        session = self._get_xenstore_session()
        session.EndSession()
        self.session = None
        pythoncom.CoUninitialize()

        self.r_terminator.shutdown(socket.SHUT_RDWR)

        self.r_terminator.close()
        self.w_terminator.close()

        self.response_packets = None
        self.r_terminator = self.w_terminator = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os.path

from setuptools import setup

README = ''
if os.path.exists('README.md'):
    with open('README.md') as readme_file:
        README = readme_file.read()

REQUIREMENTS = [
    'pywin32',
    'pyxs',
    'six',
    'WMI',
]

TEST_REQUIREMENTS = [
    'mock',
]

setup(
    name='win_pyxs',
    version='0.1.0',
    description='Windows compatibility for pyxs',
    long_description=README,
    author='Joel Noyce Barnham',
    author_email='joelnbarnham@gmail.com',
    url='https://github.com/joelnb/win-pyxs',
    packages=['win_pyxs'],
    package_dir={
        'win_pyxs': 'win_pyxs',
    },
    include_package_data=True,
    install_requires=REQUIREMENTS,
    zip_safe=True,
    keywords='pyxs xenstore xen',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    test_suite='tests',
    tests_require=TEST_REQUIREMENTS,
    extras_requires={
        ':sys_platform == "win32" and (python_version < "3.5")': [
            'backports.socketpair',
        ]
    }
)



"""
An implementation of pyxs.connection.PacketConnection for the Windows
versions of the Xen PV drivers. There are two possible connections:
XenBusConnectionWinPV which can be used with the WinPV drivers available for
modern versions of Windows & XenBusConnectionGPLPV which works with the GPLPV
drivers which were available for older versions of Windows.
"""

__all__ = ['XenBusConnectionWinPV', 'XenBusConnectionGPLPV']

from .gplpv import XenBusConnectionGPLPV
from .winpv import XenBusConnectionWinPV
